from colouration import Colour


DEFAULT_COLOUR = Colour(hexadecimal='#AAAAAA')
DEFAULT_TEXT_SIZE = 10
DEFAULT_FONT = 'helvetica'
DEFAULT_SHAPE = 'box'
DEFAULT_SHAPE_STYLE = 'rounded, filled'
DEFAULT_NODE_COLOUR_DIFFERENCE_RATIO = 0.5


class NodeStyle:
	def __init__(
			self, colour=DEFAULT_COLOUR, fill_colour=None, border_colour=None, opacity=None,
			font=DEFAULT_FONT, text_colour=None, text_size=DEFAULT_TEXT_SIZE,
			shape=DEFAULT_SHAPE, shape_style=DEFAULT_SHAPE_STYLE, lighter_fill=True
	):
		colour = colour or fill_colour or border_colour or DEFAULT_COLOUR
		colour = Colour(obj=colour)

		self._fill_colour_based_on_main_colour = False
		self._border_colour_based_on_main_colour = False
		self._text_colour_based_on_main_colour = False
		self._lighter_fill = lighter_fill

		self._colour = colour
		self._fill_colour = None
		self._border_colour = None
		self._text_colour = None

		self.fill_colour = fill_colour
		self.border_colour = border_colour
		self.text_colour = text_colour
		self.colour = colour

		self._opacity = opacity
		self._font = font
		self._text_size = text_size
		self._shape = shape
		self._shape_style = shape_style

	def complement(self, dictionary):
		"""
		:type dictionary: dict
		"""
		for key, value in dictionary.items():
			if hasattr(self, key):
				setattr(self, key, value)
			elif not key.startswith('_'):
				setattr(self, f'_{key}', value)
			else:
				raise ValueError(f'{key} is not a valid attribute of NodeStyle')

	def reset_colours(self):
		self._fill_colour = None
		self._border_colour = None
		self._text_colour = None

	def invert_colours(self):
		"""
		:rtype: NodeStyle
		"""
		result = self.copy()

		if result._fill_colour_based_on_main_colour:
			result._fill_colour = None
		else:
			result.fill_colour = result.fill_colour.invert()

		if result._text_colour_based_on_main_colour:
			result._text_colour = None
		else:
			result.text_colour = result.text_colour.invert()

		if result._border_colour_based_on_main_colour:
			result._border_colour = None
		else:
			result.border_colour = result.border_colour.invert()

		if result.colour is not None:
			result.colour = result.colour.invert()

		return result

	def copy(self):
		"""
		:rtype: NodeStyle
		"""
		result = self.__class__(
			colour=self._colour, fill_colour=self._fill_colour, border_colour=self._border_colour,
			opacity=self._opacity, font=self._font, text_colour=self._text_colour, text_size=self._text_size,
			shape=self._shape, shape_style=self._shape_style
		)
		result._fill_colour_based_on_main_colour = self._fill_colour_based_on_main_colour
		result._text_colour_based_on_main_colour = self._text_colour_based_on_main_colour
		result._border_colour_based_on_main_colour = self._border_colour_based_on_main_colour
		return result

	@property
	def colour(self):
		"""
		:rtype: Colour
		"""
		return self._colour

	@colour.setter
	def colour(self, colour):
		if isinstance(colour, str):
			colour = Colour(obj=colour)
		self._colour = colour
		if self._fill_colour is None:
			if self._lighter_fill:
				self._fill_colour = colour.lighten(ratio=DEFAULT_NODE_COLOUR_DIFFERENCE_RATIO)
			else:
				self._fill_colour = colour.darken(ratio=DEFAULT_NODE_COLOUR_DIFFERENCE_RATIO)
			self._fill_colour_based_on_main_colour = True
		if self._border_colour is None:
			if self._lighter_fill:
				self._border_colour = colour.darken(ratio=DEFAULT_NODE_COLOUR_DIFFERENCE_RATIO / 3)
			else:
				self._border_colour = colour.lighten(ratio=DEFAULT_NODE_COLOUR_DIFFERENCE_RATIO / 3)
			self._border_colour_based_on_main_colour = True
		if self._text_colour is None:
			self._text_colour = self.fill_colour.farthest_gray
			self._text_colour_based_on_main_colour = True

	@property
	def fill_colour(self):
		"""
		:rtype: Colour
		"""
		return self._fill_colour

	@fill_colour.setter
	def fill_colour(self, fill_colour):
		if fill_colour is not None:
			self._fill_colour = Colour(fill_colour)

	@property
	def border_colour(self):
		"""
		:rtype: Colour
		"""
		return self._border_colour

	@border_colour.setter
	def border_colour(self, border_colour):
		if border_colour is not None:
			self._border_colour = Colour(border_colour)

	@property
	def text_colour(self):
		"""
		:rtype: Colour
		"""
		return self._text_colour

	@text_colour.setter
	def text_colour(self, text_colour):
		if text_colour is not None:
			self._text_colour = Colour(text_colour)
			self._text_colour_based_on_main_colour = False

	def __getstate__(self):
		return {
			'_colour': self._colour,
			'_fill_colour': self._fill_colour,
			'_border_colour': self._border_colour,
			'_font': self._font,
			'_text_colour': self._text_colour,
			'_text_size': self._text_size,
			'_shape': self._shape,
			'_shape_style': self._shape_style
		}

	def __setstate__(self, state):
		self._colour = state['_colour']
		self._fill_colour = state['_fill_colour']
		self._border_colour = state['_border_colour']
		self._opacity = state['_opacity']
		self._font = state['_font']
		self._text_colour = state['_text_colour']
		self._text_size = state['_text_size']
		self._shape = state['_shape']
		self._shape_style = state['_shape_style']

	@property
	def _graphviz_dictionary(self):
		return {
			'color': self.border_colour,
			'fillcolor': self.fill_colour,
			'fontname': self._font,
			'fontcolor': self.text_colour,
			'fontsize': self._text_size,
			'opacity': self._opacity,
			'shape': self._shape,
			'style': self._shape_style
		}

	def get_graphviz_str(self):
		"""
		:rtype: str
		"""
		key_values = []
		for key, value in self._graphviz_dictionary.items():
			if isinstance(value, Colour):
				value = value.get_hexadecimal(opacity=self._opacity)

			if isinstance(key, str):
				key = key.strip('"')

			if isinstance(value, str):
				value = value.strip('"')

			if value is not None:
				key_values.append((key, value))

		return ' '.join([f'"{key}"="{value}"' for key, value in key_values])



from ._GraphObj import GraphObj
from .styling.EdgeStyle import EdgeStyle
from .Node import Node
from typing import Optional, Union, Dict, Tuple


class Edge(GraphObj):
	def __init__(
			self, graph: 'Graph', start: Node, end: Node, id: Optional[Union[str, int]] = None, 
			value: Optional[object] = None, label: Optional[str] = None, 
			tooltip: Optional[str] = None, style: Optional[EdgeStyle] = None, **kwargs
	):
		"""
		Initializes an Edge instance.

		Args:
			graph (Graph): A graph.
			start (Node): Start node.
			end (Node): End node.
			id (Optional[Union[str, int]]): An extra ID (for when there are multiple edges between the same start and end nodes).
			value (Optional[object]): Some value the edge keeps.
			label (Optional[str]): A label to show.
			tooltip (Optional[str]): Tooltip for the edge.
			style (Optional[EdgeStyle]): The style of the edge.
			**kwargs: Additional keyword arguments.
		"""

		super().__init__(
			graph=graph, id=None, value=value, label=label, tooltip=tooltip, style=style,
			**kwargs
		)
		self._start = start
		self._end = end
		self.raw_id = id
		start.append_outward_edge(edge=self)
		end.append_inward_edge(edge=self)

	@property
	def _label_converter(self):
		"""Gets the label converter for the edge."""
		return self.graph._edge_label_converter

	@property
	def style(self) -> Optional[EdgeStyle]:
		"""
		Gets the style of the edge.

		Returns:
			Optional[EdgeStyle]: The style of the edge.
		"""
		return self._style

	@style.setter
	def style(self, style: Union[EdgeStyle, Dict]):
		"""
		Sets the style of the edge.

		Args:
			style (Union[EdgeStyle, Dict]): The style to set, can be an EdgeStyle instance or a dictionary.
		
		Raises:
			RuntimeError: If overwriting the style is not allowed.
			TypeError: If the style type is not supported.
		"""
		if self._style is not None and not self.graph._style_overwrite_allowed:
			raise RuntimeError('Cannot overwrite edge style!')

		if isinstance(style, dict):
			self._style = EdgeStyle(**style)
		elif isinstance(style, EdgeStyle):
			self._style = style
		elif style is None:
			pass
		else:
			raise TypeError(f'edge style of type {type(style)} is not supported!')

	def is_similar_to(self, other: 'Edge') -> bool:
		"""
		Checks if this edge is similar to another edge.

		Args:
			other (Edge): The other edge to compare.

		Returns:
			bool: True if similar, False otherwise.
		"""
		if not super().is_similar_to(other=other):
			return False
		return (self._start.id == other._start.id and 
				self._end.id == other._end.id and 
				self.raw_id == other.raw_id)

	def __getstate__(self) -> Dict:
		"""Returns the state of the edge for pickling."""
		state = super().__getstate__()
		state.update({
			'start': None,
			'end': None,
			'id': self.raw_id
		})
		return state

	def __setstate__(self, state: Dict):
		"""Restores the state of the edge from a pickled state."""
		super().__setstate__(state=state)
		self._start = state['start']
		self._end = state['end']
		self.raw_id = state['id']

	@property
	def id(self) -> Tuple[Union[str, int]]:
		"""
		Gets the ID of the edge.

		Returns:
			Tuple[Union[str, int]]: The ID of the edge.
		
		Raises:
			ValueError: If the edge does not have a start or end.
		"""
		if self._start is None:
			raise ValueError('This Edge does not have a start!')
		elif self._end is None:
			raise ValueError('This Edge does not have an end!')

		return self.start.id, self.end.id, self.raw_id

	def __str__(self) -> str:
		"""Returns a string representation of the edge."""
		return f'Edge:{self.start}-{self.end}'

	def __repr__(self) -> str:
		"""Returns a string representation of the edge for debugging."""
		return str(self)

	def get_graphviz_str(self) -> str:
		"""
		Gets the Graphviz string representation of the edge.

		Returns:
			str: The Graphviz representation.
		"""
		parts = []
		label_or_value = self.display_label_or_value()
		style = self.style
		if label_or_value is not None:
			parts.append(f'label="{label_or_value}"')
		if self._tooltip is not None:
			parts.append(f'tooltip="{self._tooltip}"')
		if style is not None:
			parts.append(style.get_graphviz_str())

		if len(parts) > 0:
			second_part = '[' + ' '.join(parts) + ']'
		else:
			second_part = ''

		return f'"{self.start.name}" -> "{self.end.name}"' + second_part

	@property
	def start(self) -> Node:
		"""
		Gets the start node of the edge.

		Returns:
			Node: The start node.

		Raises:
			ValueError: If the edge does not have a start.
		"""
		if self._start is None:
			raise ValueError('Edge does not have a start!')
		return self._start

	@property
	def end(self) -> Node:
		"""
		Gets the end node of the edge.

		Returns:
			Node: The end node.

		Raises:
			ValueError: If the edge does not have an end.
		"""
		if self._end is None:
			raise ValueError('Edge does not have an end!')
		return self._end

	def remove_self(self):
		"""Removes the edge from its start and end nodes."""
		if self._start is None or self._end is None:
			raise ValueError('Either start or end Node is missing!')
		self.start.remove_outward_edge(edge_id=self.id)
		self.end.remove_inward_edge(edge_id=self.id)
		self._start = None
		self._end = None
		self._graph = None

	def is_in_loop(self) -> bool:
		"""Checks if the edge is in a loop."""
		return self.start.is_in_loop_with(self.end)

	# TODO: Uncomment if needed
	# @property
	# def graph_style(self):
	#     return self.graph.edge_style



from ._BasicGraph import BasicGraph
from .styling.NodeStyle import NodeStyle
from .styling.EdgeStyle import EdgeStyle
from graphviz import Source
import os
from functools import wraps
import random
from colouration import Colour
from .styling import stylize_with_pensieve, stylize_randomly


DEFAULT_BACKGROUND_COLOUR_NAME = '#FAFAFA'
DEFAULT_PAD = 0.1


class GraphWithoutDisplay(BasicGraph):
	def __init__(
			self, obj=None, strict=True, ordering=True, node_style=None, edge_style=None,
			node_label_converter=None, edge_label_converter=None,
			colour_scheme='pensieve2', background_colour=DEFAULT_BACKGROUND_COLOUR_NAME,
			font='helvetica',
			direction='LR', stylist='pensieve', label='\nPowered by Abstract', label_url='https://github.com/idin/abstract',
			tooltip=None,
			label_location='bottom', font_size=10, label_colour='deepskyblue3', label_background_colour=None,
			style_overwrite_allowed=False,
			**kwargs
	):
		"""
		:param style_overwrite_allowed: if False, a node style cannot be overwritten
		"""

		self._background_colour = None
		self._direction = direction
		self.background_colour = background_colour

		self._global_node_style_overwrite = node_style
		self._global_edge_style_overwrite = edge_style
		self._node_style_overwrites = {}
		self._edge_style_overwrites = {}
		self._node_colour_overwrites = {}
		self._edge_colour_overwrites = {}
		self._label = label
		self._label_location = label_location
		self._label_colour = label_colour
		self._label_url = label_url
		self._label_background_colour = label_background_colour

		self._tooltip = tooltip

		self._font = font
		self._font_size = font_size
		self._kwargs = kwargs

		if isinstance(obj, self.__class__):
			colour_scheme = colour_scheme or obj._colour_scheme
		else:
			try:
				obj = obj.__graph__()
			except AttributeError:
				pass

		if isinstance(obj, dict):
			if 'colour_scheme' in obj:
				colour_scheme = colour_scheme or obj['colour_scheme']

			if 'style_overwrite_allowed' in obj:
				style_overwrite_allowed = obj['style_overwrite_allowed']

			if 'stylist' in obj:
				stylist = stylist or obj['stylist']
			if stylist is None:
				stylist = 'pensieve'

		self._colour_scheme = colour_scheme
		self._stylist = stylist
		self._style_overwrite_allowed = style_overwrite_allowed

		super().__init__(
			strict=strict, ordering=ordering,
			node_label_converter=node_label_converter, edge_label_converter=edge_label_converter
		)

		if obj:
			self.append(obj=obj)

	def __getstate__(self):
		state = super().__getstate__()
		return state

	def update_nodes(self):
		if not self._nodes_have_graph:
			for node in self._nodes_dict.values():
				node._graph = self
				node.update_edges()
		self._nodes_have_graph = True

	def __setstate__(self, state):
		super().__setstate__(state=state)
		'''
		self._node_style = state['node_style']
		self._edge_style = state['edge_style']
		'''
		self._nodes_have_graph = False
		self.update_nodes()

	@wraps(BasicGraph.connect)
	def connect(self, start, end, style=None, **kwargs):
		return super().connect(start=start, end=end, style=style, **kwargs)

	@wraps(BasicGraph.add_node)
	def add_node(self, name, style=None, **kwargs):
		# if name is a Node it should be handled by the BasicGraph.add_node method
		return super().add_node(name=name, style=style, **kwargs)

	@wraps(BasicGraph.disconnect)
	def disconnect(self, edge):
		return super().disconnect(edge=edge)

	@property
	def background_colour(self):
		"""
		:rtype: Colour
		"""
		return self._background_colour

	@background_colour.setter
	def background_colour(self, background_colour):
		"""
		:type background_colour: Colour or str
		"""
		if not isinstance(background_colour, Colour):
			background_colour = Colour(obj=background_colour)
		self._background_colour = background_colour

	@property
	def node_styles(self):
		"""
		:rtype: dict[str, NodeStyle]
		"""
		return {name: node.style for name, node in self.nodes_dict.items() if node.has_style()}

	@property
	def edge_styles(self):
		"""
		:rtype: dict[str, EdgeStyle]
		"""
		return {edge.id: edge.style for node in self.nodes for edge in node.outward_edges if edge.has_style()}

	def stylize(self):
		if self._global_node_style_overwrite is not None:
			smart_global_node_style_overwrite = None
			if 'shape' in self._global_node_style_overwrite:
				if self._global_node_style_overwrite['shape'].lower().startswith('auto'):
					if max([len(str(node.name)) for node in self.nodes]) < 3:
						smart_global_node_style_overwrite = self._global_node_style_overwrite.copy()
						smart_global_node_style_overwrite['shape'] = 'circle'
			if smart_global_node_style_overwrite is None:
				smart_global_node_style_overwrite = self._global_node_style_overwrite

			for node in self.nodes:
				node.style = smart_global_node_style_overwrite

		if self._global_edge_style_overwrite is not None:
			for edge in self.edges:
				edge.style = self._global_edge_style_overwrite

		if self._stylist == 'pensieve':
			stylize_with_pensieve(graph=self)
		elif self._stylist == 'random':
			stylize_randomly(graph=self)

		for name, style in self._node_style_overwrites.items():
			self.nodes_dict[name].style.complement(style)

		for edge_id, style in self._edge_style_overwrites.items():
			self.edges_dict[edge_id].style = style

		for name, colour in self._node_colour_overwrites.items():
			node = self.nodes_dict[name]
			style = node.style.copy()
			style.reset_colours()
			style.colour = colour
			node.style = style

		for edge_id, colour in self._edge_colour_overwrites.items():
			edge = self.edges_dict[edge_id]
			_ = edge.style.colour
			style = edge.style.copy()
			style.reset_colours()
			style.colour = colour
			edge.style = style
			_ = edge.style.colour

		if self._global_node_style_overwrite is not None:
			smart_global_node_style_overwrite = None
			if 'shape' in self._global_node_style_overwrite:
				if self._global_node_style_overwrite['shape'].lower().startswith('auto'):
					if max([len(str(node.name)) for node in self.nodes]) < 3:
						smart_global_node_style_overwrite = self._global_node_style_overwrite.copy()
						smart_global_node_style_overwrite['shape'] = 'circle'
			if smart_global_node_style_overwrite is None:
				smart_global_node_style_overwrite = self._global_node_style_overwrite

			for node in self.nodes:
				node.style.complement(smart_global_node_style_overwrite)

		if self._global_edge_style_overwrite is not None:
			for edge in self.edges:
				edge.style.complement(self._global_edge_style_overwrite)

	def get_graphviz_header(self, dpi=300, direction=None, height=None, width=None, pad=DEFAULT_PAD):
		"""
		:type direction: str or NoneType
		:type dpi: int
		:type height: float or int or NoneType
		:type width: float or int or NoneType
		:type pad: int or float or NoneType
		:rtype: str
		"""
		direction = direction or self._direction

		if self._is_strict:
			first_part = 'strict digraph G {\n'
		else:
			first_part = 'digraph G{\n'

		second_part = ''

		attributes = {}
		if self._label is not None:
			if self._label.startswith('<') and self._label.endswith('>'):
				attributes['label'] = f'{self._label}'
			else:
				if self._label_background_colour is not None:
					html = f'<table border="0" cellborder="0"><tr><td bgcolor="{self._label_background_colour}" title="{self._label}">{self._label} </td></tr></table>'
					if self._label_location[0] == 't':
						attributes['label'] = f'<{html}\n>'
					else:
						attributes['label'] = f'<\n{html}>'
				else:
					attributes['label'] = f'"{self._label}"'

			if self._label_location is not None:
				attributes['labelloc'] = f'"{self._label_location[0]}"'
			if self._font_size is not None:
				attributes['fontsize'] = f'{self._font_size}'
			if self._font is not None:
				attributes['fontname'] = f'"{self._font}"'
			if self._label_colour is not None:
				attributes['fontcolor'] = f'"{self._label_colour}"'
			if self._label_url is not None:
				attributes['href'] = f'"{self._label_url}"'
				attributes['target'] = '"_blank"'

		if self._tooltip is not None:
			attributes['tooltip'] = f'"{self._tooltip}"'

		#if dpi is not None:
		#	attributes['dpi'] = f'{dpi}'
		for key, value in self._kwargs.items():
			if isinstance(value, int):
				attributes[key] = value
			else:
				attributes[key] = f'"{value}"'

		if len(attributes) > 0:
			second_part += '\t graph [' + ', '.join([f'{key} = {value}' for key, value in attributes.items()]) + ' ];\n'

		second_part += f'\tbgcolor="{self.background_colour.hexadecimal}";\n'

		if pad is not None:
			second_part += f'\tpad="{pad}";\n'

		if not self._ordering:
			second_part += '\tordering=out;\n'

		if height is not None and width is not None:
			second_part += f'\tsize="{width},{height}!";\n'
			second_part += '\tratio="fill";\n'
		elif height is not None:
			second_part += f'\tsize="{height}!";\n'
			second_part += '\tratio="fill";\n'
		elif width is not None:
			second_part += f'\tsize="{width}!";\n'
			second_part += '\tratio="fill";\n'

		direction = direction.upper()
		if direction != 'TB':
			third_part = f'\trankdir="{direction}"'
		else:
			third_part = ''

		return first_part + second_part + third_part

	def get_graphviz_str(self, direction=None, dpi=300, height=None, width=None, pad=DEFAULT_PAD):
		"""
		:type direction: str or NoneType
		:type dpi: int or NoneType
		:type height: float or int or NoneType
		:type width: float or int or NoneType
		:type pad: int or float or NoneType
		:rtype: str
		"""
		direction = direction or self._direction

		nodes_str = '\t{\n\t\t' + '\n\t\t'.join([node.get_graphviz_str() for node in self.nodes_dict.values()]) + '\n\t}\n'
		edges_str = '\t' + '\n\t'.join([edge.get_graphviz_str() for edge in self.edges]) + '\n'
		header = self.get_graphviz_header(direction=direction, dpi=dpi, height=height, width=width, pad=pad)
		return header + nodes_str + edges_str + '}'

	def append(self, obj):
		"""
		adds nodes and edges from an object's __graph__() method which is essentially just a dictionary
		:type obj: Object
		:rtype: Graph
		"""
		if isinstance(obj, (list, tuple)):
			for item in obj:
				self.append(obj=item)
			return self

		else:
			try:
				dictionary = obj.__graph__()
			except AttributeError:
				dictionary = obj

			if not isinstance(dictionary, dict):
				raise TypeError(f'{type(obj)} cannot be converted to a Graph. It needs to have a __graph__() method.')  # ToDo cite proper link for explanation

			if 'strict' in dictionary:
				self._is_strict = dictionary['strict']

			if 'ordering' in dictionary:
				self._ordering = dictionary['ordering']

			if 'label' in dictionary:
				self._label = '\n' + dictionary['label']

			if 'tooltip' in dictionary:
				self._tooltip = dictionary['tooltip']

			if 'label_url' in dictionary:
				self._label_url = dictionary['label_url']

			if 'global_node_style' in dictionary:
				self._global_node_style_overwrite = dictionary['global_node_style']
			elif 'graph_node_style' in dictionary:
				self._global_node_style_overwrite = dictionary['graph_node_style']

			if 'global_edge_style' in dictionary:
				self._global_edge_style_overwrite = dictionary['global_edge_style']
			elif 'graph_edge_style' in dictionary:
				self._global_edge_style_overwrite = dictionary['graph_edge_style']

			if 'direction' in dictionary:
				self._direction = dictionary['direction']

			if isinstance(dictionary['nodes'], dict):
				for node_name, node_dict in dictionary['nodes'].items():
					self.add_node(name=node_name, **node_dict)
			else:
				for node_name in dictionary['nodes']:
					self.add_node(name=node_name)

			edges_dict = {}
			for parent_child_edge_dict in dictionary['edges']:

				if len(parent_child_edge_dict) == 2:
					parent, child = parent_child_edge_dict
					edge = self.connect(start=parent, end=child)
					edges_dict[(parent, child)] = edge.id

				elif len(parent_child_edge_dict) == 3:
					parent, child, edge_dict = parent_child_edge_dict
					edge = self.connect(start=parent, end=child, **edge_dict)
					edges_dict[(parent, child)] = edge.id

				elif len(parent_child_edge_dict) < 2:
					raise ValueError('Too few objects in an edge definition!')
				else:
					raise ValueError('Too many objects in an edge definition!')

			if 'node_styles' in dictionary:
				for name, node_style in dictionary['node_styles'].items():
					self._node_style_overwrites[name] = node_style

			if 'edge_styles' in dictionary:
				for parent_child_edge_id, edge_style in dictionary['edge_styles'].items():
					edge_id = edges_dict[parent_child_edge_id]
					self._edge_colour_overwrites[edge_id] = edge_style

			if 'node_colours' in dictionary:
				for name, colour in dictionary['node_colours'].items():
					self._node_colour_overwrites[name] = colour

			if 'edge_colours' in dictionary:
				for parent_child_edge_id, colour in dictionary['edge_colours'].items():
					edge_id = edges_dict[parent_child_edge_id]
					self._edge_colour_overwrites[edge_id] = colour

			return self

	@classmethod
	def from_dict(cls, obj):
		"""
		:param obj:
		:rtype: Graph
		"""
		graph = cls()
		graph.append(obj=obj)
		return graph

	def __graph__(self):
		"""
		:rtype: dict
		"""
		nodes_dict = {}
		edges_list = []
		for node in self.nodes:
			nodes_dict[node.id] = {
				'label': node.display_label_or_value,
				'value': node.value,
				'style': node.style
			}

			for edge in node.outward_edges:
				edges_list.append([
					edge.start.id,
					edge.end.id,
					{
						'id': edge.raw_id,
						'label': edge.raw_label,
						'value': edge.value,
						'style': edge.style
					}
				])

		return {
			'strict': self._is_strict,
			'ordering': self._ordering,
			'colour_scheme': self._colour_scheme,
			'label': self._label,
			'label_url': self._label_url,
			'node_styles': self._node_style_overwrites,
			'edge_styles': self._edge_style_overwrites,
			'global_node_style': self._global_node_style_overwrite,
			'global_edge_style': self._global_edge_style_overwrite,
			'node_colours': self._node_colour_overwrites,
			'nodes': nodes_dict,
			'edges': edges_list
		}

	@classmethod
	def random(
			cls, num_nodes, cycle=False, start_index=1, connection_probability=0.5,
			ordering=True, seed=None
	):
		"""
		creates a random graph
		:param int num_nodes:
		:param bool cycle:
		:param int start_index:
		:param float connection_probability:
		:param bool ordering:
		:rtype: BasicGraph
		"""
		graph = cls(strict=True, ordering=ordering)
		connection_probability = min(1.0, max(0.0, connection_probability))
		node_names = [i + start_index for i in range(num_nodes)]
		for n in node_names:
			graph.add_node(name=str(n))

		if seed is not None:
			random.seed(seed)
		for n1 in node_names:
			for n2 in node_names:
				if random.uniform(0, 1) < connection_probability:
					if cycle:
						graph.connect(start=str(n1), end=str(n2))
					elif n1 < n2:
						graph.connect(start=str(n1), end=str(n2))

		return graph

	def __add__(self, other):
		"""
		combines two graphs
		:type other: Graph
		:rtype: Graph
		"""
		return self.add(other=other)

	def add(self, other, add_values_function=None):
		"""
		combines two graphs
		:type self: Graph
		:type other: Graph
		:type add_values_function: callable
		:rtype: Graph
		"""

		d1 = self.__graph__()
		d2 = other.__graph__()

		node_styles = d1['node_styles']
		node_styles.update(d2['node_styles'])

		edge_styles = d1['edge_styles']
		edge_styles.update(d2['edge_styles'])

		if add_values_function is None:
			def add_values_function(x, y):
				if x is None or y is None:
					return x or y
				else:
					return x + y

		nodes = {}
		for the_id in set(d1['nodes'].keys()).union(d2['nodes'].keys()):
			if the_id in d1['nodes'] and the_id in d2['nodes']:
				node1 = d1['nodes'][the_id]
				node2 = d2['nodes'][the_id]
				n = {}
				for key in set(node1.keys()).union(node2.keys()):
					if key in node1 and key in node2:
						if key == 'value':
							n[key] = add_values_function(node2[key], node2[key])
						else:
							n[key] = node1[key] or node2[key]
					elif key in node1:
						n[key] = node1[key]
					else:
						n[key] = node2[key]
				nodes[the_id] = n
			elif the_id in d1['nodes']:
				nodes[the_id] = d1['nodes'][the_id]
			else:
				nodes[the_id] = d2['nodes'][the_id]

		edges = {}
		for edge2 in d1['edges'] + d2['edges']:
			start2, end2, edge2_dict = edge2
			id2 = edge2_dict['id']
			if (start2, end2, id2) in edges:
				edge1 = edges[(start2, end2, id2)]
				edge1_dict = edge1[2]
				for key in set(edge1_dict.keys()).union(set(edge2_dict.keys())):
					if key in edge1_dict and key in edge2_dict:
						if key == 'value':
							edge1_dict[key] = add_values_function(edge1_dict[key], edge2_dict[key])
						else:
							edge1_dict[key] = edge1_dict[key] or edge2_dict[key]
					elif key in edge2_dict:
						edge1_dict[key] = edge2_dict[key]
			else:
				edges[(start2, end2, id2)] = start2, end2, edge2_dict
		edges = edges.values()

		result_representation = {
			'strict': d1['strict'],
			'ordering': d1['ordering'],
			'node_styles': node_styles,
			'edge_styles': edge_styles,
			'nodes': nodes,
			'edges': edges
		}
		result = self.__class__(obj=result_representation)

		return result

	def render(
			self, path=None, view=True, direction=None, height=None, width=None, dpi=300, pad=DEFAULT_PAD
	):
		"""
		:type direction: NoneType or str
		:type dpi: NoneType or int
		:type height: NoneType of int or float
		:type width: NoneType or int or float
		:type pad: NoneType or int or float
		:type output_format: str or NoneType
		:rtype: Source
		"""
		direction = direction or self._direction

		self.stylize()

		if path is None:
			return self.get_graphviz_source(direction=direction, pad=pad, dpi=None)# Source(source=self.get_graphviz_str(direction=direction, pad=pad, dpi=None))
		else:
			filename, file_extension = os.path.splitext(path)
			output_format = file_extension.lstrip('.')
			to_save = self.get_graphviz_source(
				direction=direction, pad=pad, dpi=dpi, height=height, width=width, output_format=output_format
			)
			to_save.render(filename=filename, view=view)
			return self.get_graphviz_source(direction=direction, pad=pad, dpi=None)

	"""
	def display_html(self, direction=None, pad=None, echo_errors=False, **kwargs):
		try:
			return displayHTML(self.get_svg(direction=direction, pad=pad, **kwargs))
		except Exception as e:
			if echo_errors:
				print('Could not use displayHTML!')
				print(e)

			html = self.get_html(direction=direction, pad=pad, **kwargs)
			try:
				return display(html)
			except Exception as e:
				if echo_errors:
					print('Coult not use display!')
					print(e)
				from IPython.core.display import display as ipython_display
				return ipython_display(html)
	"""

	def get_graphviz_source(self, direction=None, dpi=300, height=None, width=None, pad=DEFAULT_PAD, output_format=None):
		"""
		:type direction: NoneType or str
		:type dpi: NoneType or int
		:type height: NoneType of int or float
		:type width: NoneType or int or float
		:type pad: NoneType or int or float
		:type output_format: str or NoneType
		:rtype: Source
		"""
		if height is None and width is None:
			return Source(source=self.get_graphviz_str(direction=direction, pad=pad, dpi=None))
		else:
			graphviz_str_to_save = self.get_graphviz_str(
				direction=direction, height=height, width=width, pad=pad, dpi=dpi
			)
			return Source(source=graphviz_str_to_save, format=output_format)



from .Graph import Graph
from .Node import Node
from .Edge import Edge
from .get_ancestors import get_ancestors
from .get_descendants import get_descendants

__all__ = ['Graph', 'Node', 'Edge', 'get_ancestors', 'get_descendants']



from .EdgeStyle import EdgeStyle


def stylize_edges_based_on_nodes(graph, edge_style=None, edge_darkness_ratio=0.1):
	for edge in graph.edges:
		if edge.style is None:
			colour = edge.start.style.colour.darken(ratio=edge_darkness_ratio)

			if edge_style is None:
				edge_style = EdgeStyle(colour=colour)
			else:
				edge_style = edge_style.copy()
				edge_style.reset_colours()
				edge_style.colour = colour

			if edge.value is not None:
				edge_style._arrow_size *= edge.value

			edge.style = edge_style


from ._GraphObj import GraphObj
from .styling.NodeStyle import NodeStyle
from typing import Optional, Union, List, Dict


CORNER = u'\u2514'
TWO_WAY = u'\u251C'
HORIZONTAL = u'\u2500'
VERTICAL = u'\u2502'


class Node(GraphObj):
	def __init__(
			self, graph: 'Graph', name: str, value: Optional[object] = None, 
			label: Optional[str] = None, tooltip: Optional[str] = None, 
			style: Optional[NodeStyle] = None, index: int = 0, **kwargs
	):
		"""
		Initializes a Node instance.

		Args:
			graph (Graph): The graph this node belongs to.
			name (str): Name of the node (code friendly, no weird characters, etc.) it should be unique in the graph.
			value (Optional[object]): Any object stored in the node.
			label (Optional[str]): A more presentable/user-friendly name.
			tooltip (Optional[str]): Tooltip for the node.
			style (Optional[NodeStyle]): Style for the node.
			index (int): Index of the node.
			**kwargs: Additional keyword arguments.
		"""

		super().__init__(
			graph=graph, id=name, value=value, label=label, tooltip=tooltip, style=style,
			# additional_styling_function=additional_styling_function,
			**kwargs
		)
		self._outward_edges_dict = dict()
		self._inward_edges_dict = dict()
		self._outward_edges_have_start_node = True
		self._inward_edges_have_end_node = True
		self._index = index
		graph._nodes_dict[name] = self

	# make node hashable
	def __hash__(self):
		return hash(self.id)

	@property
	def style(self) -> Optional[NodeStyle]:
		"""
		Gets the style of the node.

		Returns:
			Optional[NodeStyle]: The style of the node.
		"""
		return self._style

	@style.setter
	def style(self, style: Union[NodeStyle, Dict]):
		"""
		Sets the style of the node.

		Args:
			style (Union[NodeStyle, Dict]): The style to set, can be a NodeStyle instance or a dictionary.
		
		Raises:
			RuntimeError: If overwriting the style is not allowed.
			TypeError: If the style type is not supported.
		"""
		if self._style is not None and not self.graph._style_overwrite_allowed:
			raise RuntimeError('Cannot overwrite node style!')

		if isinstance(style, dict):
			self._style = NodeStyle(**style)
		elif isinstance(style, NodeStyle):
			self._style = style
		elif style is None:
			pass
		else:
			raise TypeError(f'node style of type {type(style)} is not supported!')

	@property
	def index(self) -> int:
		"""Gets the index of the node."""
		return self._index

	def __getstate__(self) -> Dict:
		"""Returns the state of the node for pickling."""
		state = super().__getstate__()
		state.update({
			'outward_edges_dict': self._outward_edges_dict,
			'inward_edges_dict': self._inward_edges_dict,
			'index': self._index
		})
		return state

	def __setstate__(self, state: Dict):
		"""Restores the state of the node from a pickled state."""
		super().__setstate__(state=state)
		self._outward_edges_dict = state['outward_edges_dict']
		self._inward_edges_dict = state['inward_edges_dict']
		self._outward_edges_have_start_node = False
		self._inward_edges_have_end_node = False
		self._index = state['index']

	def is_similar_to(self, other: 'Node') -> bool:
		"""
		Checks if this node is similar to another node.

		Args:
			other (Node): The other node to compare.

		Returns:
			bool: True if similar, False otherwise.
		"""
		if not super().is_similar_to(other=other):
			return False
		if len(self._outward_edges_dict) != len(other._outward_edges_dict):
			return False
		if len(self._inward_edges_dict) != len(other._inward_edges_dict):
			return False
		for id, outward_edge in self._outward_edges_dict.items():
			other_outward_edge = other._outward_edges_dict[id]
			if not outward_edge.is_similar_to(other=other_outward_edge):
				return False
		for id, inward_edge in self._inward_edges_dict.items():
			other_inward_edge = other._inward_edges_dict[id]
			if not inward_edge.is_similar_to(other=other_inward_edge):
				return False
		return True

	def update_edges(self):
		"""Updates the edges of the node."""
		if not self._outward_edges_have_start_node:
			for outward_edge in self._outward_edges_dict.values():
				outward_edge._start = self
				outward_edge._graph = self.graph
		self._outward_edges_have_start_node = True
		if not self._inward_edges_have_end_node:
			for inward_edge in self._inward_edges_dict.values():
				inward_edge._end = self
				inward_edge._graph = self.graph
		self._inward_edges_have_end_node = True

	@property
	def outward_edges_dict(self) -> Dict[str, 'GraphObj']:
		"""Gets the outward edges dictionary."""
		return self._outward_edges_dict

	@property
	def inward_edges_dict(self) -> Dict[str, 'GraphObj']:
		"""Gets the inward edges dictionary."""
		return self._inward_edges_dict

	@property
	def name(self) -> str:
		"""Gets the name of the node."""
		return self.id

	@property
	def outward_edge_ids(self) -> List[str]:
		"""Gets the outward edge IDs."""
		return list(self.outward_edges_dict.keys())

	@property
	def inward_edge_ids(self) -> List[str]:
		"""Gets the inward edge IDs."""
		return list(self.inward_edges_dict.keys())

	def get_outward_edge(self, id: str) -> 'GraphObj':
		"""
		Gets an outward edge by ID.

		Args:
			id (str): The ID of the outward edge.

		Returns:
			GraphObj: The outward edge.
		"""
		return self.outward_edges_dict[id]

	def get_inward_edge(self, id: str) -> 'GraphObj':
		"""
		Gets an inward edge by ID.

		Args:
			id (str): The ID of the inward edge.

		Returns:
			GraphObj: The inward edge.
		"""
		return self.inward_edges_dict[id]

	def append_outward_edge(self, edge: 'GraphObj'):
		"""
		Appends an outward edge to the node.

		Args:
			edge (GraphObj): The edge to append.
		"""
		self.outward_edges_dict[edge.id] = edge

	def append_inward_edge(self, edge: 'GraphObj'):
		"""
		Appends an inward edge to the node.

		Args:
			edge (GraphObj): The edge to append.
		"""
		self.inward_edges_dict[edge.id] = edge

	def remove_outward_edge(self, edge_id: str):
		"""Removes an outward edge by ID."""
		del self.outward_edges_dict[edge_id]

	def remove_inward_edge(self, edge_id: str):
		"""Removes an inward edge by ID."""
		del self.inward_edges_dict[edge_id]

	@property
	def label(self) -> str:
		"""Gets the label of the node."""
		if self._label:
			result = str(self._label)
		else:
			result = str(self.id)
		return result.replace('"', '\\"')

	@label.setter
	def label(self, label: str):
		"""Sets the label of the node."""
		self._label = label

	@property
	def _label_converter(self):
		"""Gets the label converter for the node."""
		return self.graph._node_label_converter

	def __str__(self) -> str:
		"""Returns a string representation of the node."""
		return f'Node:{self.id}'

	def __repr__(self) -> str:
		"""Returns a string representation of the node for debugging."""
		return str(self)

	def __lt__(self, other: 'Node') -> bool:
		"""Less than comparison for sorting nodes."""
		return (self.index, self.name) < (other.index, other.name)

	def __gt__(self, other: 'Node') -> bool:
		"""Greater than comparison for sorting nodes."""
		return (self.index, self.name) > (other.index, other.name)

	def __le__(self, other: 'Node') -> bool:
		"""Less than or equal comparison for sorting nodes."""
		return (self.index, self.name) <= (other.index, other.name)

	def __ge__(self, other: 'Node') -> bool:
		"""Greater than or equal comparison for sorting nodes."""
		return (self.index, self.name) >= (other.index, other.name)

	def __eq__(self, other: 'Node') -> bool:
		"""Equality comparison for nodes."""
		if not isinstance(other, Node):
			raise TypeError(f'node of type {type(other)} is not supported!')
		return (self.index, self.name) == (other.index, other.name)

	def __ne__(self, other: 'Node') -> bool:
		"""Inequality comparison for nodes."""
		if not isinstance(other, Node):
			raise TypeError(f'node of type {type(other)} is not supported!')
		return (self.index, self.name) != (other.index, other.name)

	def get_tree_str(self, indentation: str = '', already_added: Optional[List['Node']] = None) -> str:
		"""
		Gets a string representation of the node's tree structure.

		Args:
			indentation (str): The indentation for the tree structure.
			already_added (Optional[List[Node]]): Nodes that have already been added to avoid cycles.

		Returns:
			str: The tree structure as a string.
		"""
		if already_added is None:
			already_added = []

		tree_string = self.label

		if self in already_added:
			tree_string += '*\n'

		else:
			tree_string += '\n'
			num_children = len(self.children)

			for child_num, child in enumerate(self.children):

				already_added.append(self)
				if child_num+1 == num_children:
					tree_string += indentation + CORNER + ' ' + child.get_tree_str(
						indentation=indentation + '  ', already_added=already_added
					)

				else:
					tree_string += indentation + TWO_WAY + ' ' + child.get_tree_str(
						indentation=indentation + VERTICAL + ' ', already_added=already_added
					)

		return tree_string

	def get_graphviz_str(self) -> str:
		"""
		Gets the Graphviz string representation of the node.

		Returns:
			str: The Graphviz representation.
		"""

		parts = [f'label="{self.display_label()}"']

		if self._tooltip is not None:
			parts.append(f'tooltip="{self._tooltip}"')

		style = self.style
		if style is not None:
			parts.append(style.get_graphviz_str())

		return f'"{self.id}" [' + ' '.join(parts) + ']'

	def connect_to(self, node: 'Node', **kwargs) -> 'GraphObj':
		"""
		Connects this node to another node.

		Args:
			node (Node): The node to connect to.
			**kwargs: Additional keyword arguments.

		Returns:
			GraphObj: The resulting graph object.
		"""
		return self.graph.connect(start=self, end=node, **kwargs)

	def connect_from(self, node: 'Node', **kwargs) -> 'GraphObj':
		"""
		Connects another node to this node.

		Args:
			node (Node): The node to connect from.
			**kwargs: Additional keyword arguments.

		Returns:
			GraphObj: The resulting graph object.
		"""
		return self.graph.connect(start=node, end=self, **kwargs)

	@property
	def parents(self) -> List['Node']:
		"""Gets the parents of the node."""
		return self.graph.get_parents(node=self)
	
	@property
	def siblings(self) -> List['Node']:
		"""Gets the siblings of the node."""
		siblings = {}
		for parent in self.parents:
			for child in parent.children:
				if child.id not in siblings and child != self:
					siblings[child.id] = child
		return list(siblings.values())
	
	@property
	def co_parents(self) -> List['Node']:
		"""Gets the co-parents of the node. (spouses)"""
		co_parents = {}
		for child in self.children:
			for parent in child.parents:
				if parent.id not in co_parents and parent != self:
					co_parents[parent.id] = parent
		return list(co_parents.values())

	def has_parents(self) -> List['Node']:
		"""Checks if the node has parents."""
		return self.graph.get_parents(node=self)

	def get_child_rank(self, parent: 'Node') -> int:
		"""
		Returns the rank of this node in terms of its index relative to its siblings.

		Args:
			parent (Node): The parent node.

		Returns:
			int: The rank of this node.

		Raises:
			ValueError: If this node is not a child of the parent.
		"""
		siblings = parent.children
		if self not in siblings:
			raise ValueError(f'{self} is not a child of {parent}')
		return sorted(siblings).index(self)

	@property
	def children(self) -> List['Node']:
		"""Gets the children of the node."""
		return self.graph.get_children(node=self)

	def has_children(self) -> bool:
		"""Checks if the node has children."""
		return self.graph.has_children(node=self)

	@property
	def ancestors(self) -> List['Node']:
		"""Gets the ancestors of the node."""
		return self.graph.get_ancestors(node=self, distance=False)

	@property
	def descendants(self) -> List['Node']:
		"""Gets the descendants of the node."""
		return self.graph.get_descendants(node=self, distance=False)

	@property
	def num_descendants(self) -> int:
		"""Gets the number of descendants of the node."""
		return len(self.descendants)

	@property
	def outward_edges(self) -> List['Edge']:
		"""Gets the outward edges of the node."""
		result = []
		for edge in self._outward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		return result

	@property
	def inward_edges(self) -> List['Edge']:
		"""Gets the inward edges of the node."""
		result = []
		for edge in self._inward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		return result

	@property
	def edges(self) -> List['Edge']:
		"""Gets all edges of the node."""
		result = []
		for edge in self.outward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		for edge in self.inward_edges_dict.values():
			if edge not in result:
				result.append(edge)
		return result

	def remove_edges(self):
		"""Removes all edges from the node."""
		for edge in self.edges:
			edge.remove_self()

	@property
	def num_outward_edges(self) -> int:
		"""Gets the number of outward edges."""
		return len(self.outward_edges_dict)

	def is_in_loop(self) -> bool:
		"""Checks if the node is in a loop."""
		return self.graph.is_node_in_loop(node=self)

	def is_in_loop_with(self, other: 'Node') -> bool:
		"""Checks if this node is in a loop with another node."""
		return self.graph.are_nodes_in_same_loop(node1=self, node2=other)
